:- ensure_loaded('utils.pl').

parse(File, Ast) :- readFile(File, Content), phrase(start(Ast), Content).

readFile(File, Out) :-
    open(File, read, In),
    get_char(In, Char1),
    process_stream(Char1, In, Out),
    close(In).
 
process_stream(end_of_file, _, [])    :- !.
process_stream(Char, In, [Code|Rest]) :-
    char_code(Char, Code),
    get_char(In, Char2),
    process_stream(Char2, In, Rest).

start(start(Mod, Body)) --> optional(mod, Mod), body(Body).

mod(mod(ModId, Exports)) -->
    ws(module), name(ModId), optional(exports, Exports), ws(where).

exports(exports(Exports)) -->
    exprBracket(lparen, rparen, gcdStar, [ws(comma), export, Exports]).

export(export(IsModule, Id)) -->
    ( ws(module), {IsModule = true}
    ; {IsModule = false}
    ), name(Id).

% module's body
body(body(Impdecls, DS)) -->
    gcdStar(impdecl, Impdecls), gcdStar(declaration, DS), whitespace.

impdecl(impdecl(IsQualified, ModId, AsModId, Impspec)) --> ws(import),
    ( ws(qualified), {IsQualified = true}
    ; {IsQualified = false}
    ), fullyQualifiedName(ModId),
    ( ws(as), fullyQualifiedName(AsModId)
    ; {AsModId = empty}
    ), optional(impspec, Impspec).

impspec(impspec(Hiding, Imports)) -->
    ( ws(hiding), {Hiding = true}
    ; {Hiding = false}
    ), exprBracket(lparen, rparen, gcdStar, [ws(comma), name, Imports]).
    
declaration(declaration(D)) -->
    ( binding(D)
    ; featureDef(D)
    ; typeDef(D)
    ).

featureDef(featureDef(NameExType, FeatureBody)) -->
    nameExType(NameExType), featureBody(FeatureBody).

inlineFeatureDef(inlineFeatureDef(NameExType, Card, FeatureBody)) -->
    optBracket(lparen, rparen, nameExType, [NameExType]),
    card(lsqbracket, dots, rsqbracket, Card),
    featureBody(FeatureBody).

nameExType(nameExType(Name, Type)) --> name(Name), exType(Type).

featureBody(featureBody(CodeMapping, Subfeatures, Facts, Local)) -->
%    gcdStar(codeMapping, CodeMapping),
    optional(codeMapping, CodeMapping),
    optional(subfeatures, Subfeatures),
    optional(facts, Facts),
    optional(local, Local).
    
identifier(AI) --> isFSymf(I), takeWhile(isFSym, Is),
    {not(restricted([I|Is])), atom_codes(AI, [I|Is])}.

name(Name) --> ws(identifier(Name)).

fullyQualifiedName(fullyQualifiedName(Path, Name)) -->
    ws(gcdPlus(dot, identifier, Names)), {init(Names, Path), last(Names, Name)}.

type(type(Type)) --> ws(colon), fullyQualifiedName(Type).

inheritance(inheritance(Type)) -->
    ws(extends), fullyQualifiedName(Type).

exType(exType(I, T)) --> optional(inheritance, I), optional(type, T).

nameType(nameType(Name, Type)) --> whitespace,
    ( usefeature, fullyQualifiedName(Type), {Name = empty}
    ; name(Name), ws(colon), ws(usefeature),
      optBracket(lparen, rparen, fullyQualifiedName, [Type])
    ).

% interpreted outside
codeMapping(pattern(Mapper, Mapping, MappingParam)) -->
    ws(lanbracket), optional(mapper, Mapper),
    %conditionalEat(isNotEoc),
    name(Mapping),
    optional(mappingParam, MappingParam),
    ws(ranbracket).

mappingParam(AParam) --> whitespace, ":", whitespace, sLiteral(sLiteral(Param)),
    {atom_codes(AParam, Param)}.

mapper(mapper(M)) --> ws(dollar), identifier(M).

subfeatures(subfeatures(Subfeatures)) -->
    exprBracket(lcubracket, rcubracket, gcdStar,
    [ws(comma), subfeature, Subfeatures]).

subfeature(subfeature(Subfeature, Card, Facts)) -->
    ( ( optBracket(lparen, rparen, inlineFeatureDef, [Subfeature]),
        {Subfeature = inlineFeatureDef(_, Card, _)}
      ; optBracket(lparen, rparen, feature, [Subfeature]),
        card(lsqbracket, dots, rsqbracket, Card)
      )
    ; group(Subfeature), {Card = card(false, 1, 1)}
    ), optional(facts, Facts).

exprBracket(LB, RB, Fun, Params) --> ws(LB), callFun(Fun, Params), ws(RB).

optBracket(LB, RB, Fun, Params) -->
    ( callFun(Fun, Params)
    ; exprBracket(LB, RB, Fun, Params)
    ).

feature(feature(NameType, Facts)) -->
    nameType(NameType), optional(facts, Facts).

group(group(Card, Subfeatures)) -->
    card(lanbracket, dash, ranbracket, Card),
    exprBracket(lparen, rparen, gcdPlus, [bar, subfeature, Subfeatures]).

card(Lparen, Delim, Rparen, card(E, Min, Max)) --> whitespace,
    ( bang, {E = true}
    ; {E = false}
    ), whitespace,
    ( callFun(Lparen), ws(number(Min)), ws(callFun(Delim)), ws(number(Max)),
      ws(callFun(Rparen))
    ; asterisk, {Min = 0, Max = *}
    ; plus,     {Min = 1, Max = *}
    ; question, {Min = 0, Max = 1}
    ;           {Min = 1, Max = 1}
    ).

local(local(Local)) -->
    ws(where), exprBracket(lcubracket, rcubracket, body, [Local]).

binding(binding(Name, D)) --> name(Name), ws(equals),
    ( fullyQualifiedName(D)
    ; featureDef(D)
    ).

typeDef(typeDef(Name, Enums)) -->
    name(Name), ws(equals), gcdPlus(ws(bar), name, Enums).

facts(facts(Facts)) --> exprBracket(lsqbracket, rsqbracket, gcdStar,
    [ws(comma), fact, Facts]).                                

fact(fact(LImpliesExprs)) --> binOp(lIff, iff, lImpliesExpr, LImpliesExprs).

lImpliesExpr(LIfThenElseExprs) -->
    binOp(lImplies, implies, lIfThenElseExpr, LIfThenElseExprs).

lIfThenElseExpr(lIfThenElse(LOrExprs, LThenElseExpr)) -->
    lOrExpr(LOrExprs), optional(lThenElseExpr, LThenElseExpr).

lThenElseExpr(lThenElse(LThenOrExprs, LElseOrExprs))  -->
    ws(lThen), lOrExpr(LThenOrExprs), ws(lElse), lOrExpr(LElseOrExprs).

lOrExpr(LXorExprs)        --> binOp(lOr, or, lXorExpr, LXorExprs).

lXorExpr(LTerms)          --> binOp(lXor, xor, lTerm, LTerms).

lTerm(LNotFactors)        --> binOp(lAnd, and, lNotFactor, LNotFactors).

lNotFactor(not(LFactor))  --> ws(lNot), lFactor(LFactor).
lNotFactor(LFactor)       --> lFactor(LFactor).

lFactor(lFactor(LFactor)) -->
    ( exprBracket(lparen, rparen, fact, [LFactor])
    ; aRelation(LFactor)
    ; sRelation(LFactor)
    ; lLit(LFactor)
    ).

lLit(lLit(Quant, Elems, F)) -->
    optional(quantifier, Quant), elements(Elems), ws(bar), fact(F).
lLit(lLit(Lit)) -->
    ( fullyQualifiedName(Lit)
    ; ws(lTrue)  -> {Lit = true}
    ; ws(lFalse) -> {Lit = false}
    ).

binOp(Op, OpName, Expr, binOp(OpName, Exprs)) --> gcdPlus(ws(Op), Expr, Exprs).

sRelation(sRelation(Rel)) --> relation(sExpr, [eq, neq, subset], Rel).

aRelation(aRelation(Rel)) -->
    relation(aExpr, [eq, neq, less, lesseq, greater, greatereq], Rel).

relation(Type, Ops, relation(Expr1, Op, Expr2)) -->
    {CExpr1 =.. [Type, Expr1], CExpr2 =.. [Type, Expr2]},
    CExpr1, whitespace, {member(Op, Ops)}, Op, CExpr2.

aExpr(aExpr(ATerms)) --> gcdPlus(addop, _, aTerm, ATerms).

aTerm(aTerm(SF, AFactors)) -->
    signedFactor(SF),
    ( mulop(M), gcdStar(mulop, _, aFactor, AFacts),
      {AFactors = [M|AFacts]}
    ; {AFactors = []}
    ).

addop(plus)  --> ws(plus).
addop(minus) --> ws(minus).

mulop(mul)   --> whitespace, (asterisk ; uMultiplicationSign).
mulop(div)   --> whitespace, (slash ; uDivisionSign ; uDivisionSlash).

signedFactor(signedFactor(AFactor)) --> ws(minus), aFactor(AFactor).
signedFactor(AFactor) --> aFactor(AFactor).

aFactor(aFactor(AF)) --> whitespace,
    ( number(AF)
    ; exprBracket(lparen, rparen, aExpr, [AF])
    ; hash, sExpr(AF)
    ; value(AF)
    ; sLiteral(AF)
    ).

sExpr(setExpr(STerms)) --> gcdPlus(saddop, _, sTerm, STerms).

saddop(union) --> whitespace, (plus ; uUnion).
saddop(minus) --> whitespace, (minus ; uSetMinus).

sTerm(sTerm(SFactors)) --> gcdPlus(smulop, _, sFactor, SFactors).

smulop(intersection) --> whitespace, (ampersand ; uIntersection).

sFactor(sFactor(SF)) -->
    ( set(SF)
    ; exprBracket(lparen, rparen, sExpr, [SF])
    ).

set(set(SetDef)) -->
    exprBracket(lcubracket, rcubracket, optional, [setDef, SetDef]).

setDef(setDef(Elems, Preds)) -->
    elements(Elems), optional(predicates, Preds).

elements(elements(Name, Set)) --> name(Name), ws(in), fullyQualifiedName(Set).

predicates(predicates(Facts)) --> ws(bar), ws(gcdPlus(ws(comma), fact, Facts)).

quantifier(quantifier(Q)) --> whitespace,
    ( all  -> {Q = all}
    ; some -> {Q = some}
    ; no   -> {Q = no}
    ; lone -> {Q = lone}
    ; one  -> {Q = one}
    ).

value(value(ValIds)) --> ws(gcdPlus(dot, valId, ValIds)).

valId(valId(GetVal, Id)) -->
    ( at, {GetVal = true}
    ; {GetVal = false}
    ),
    ( exprBracket(lparen, rparen, fullyQualifiedName, [Id])
    ; identifier(Id)
    ).

sLiteral(sLiteral(CS)) --> dquote, gcdStar(sChar, CS), dquote.

sChar(Char) --> [Char], {not(isDQuote(Char)), not(isBackslash(Char))}.
sChar(ES)   --> escapeSeq(ES).

escapeSeq(escape(Char)) --> backslash, whitespace, [Char],
    {map(char_code, ['b', 't', 'n', 'f', 'r', '\"', '\'', '\\'], CS),
     member(Char, CS)}.

number(number(N)) -->
    ( asterisk, {N = '*'}
    ; number(0, N)
    ).
number(N0, N) -->
    [Char],
    {0'0 =< Char}, % 'this comment fixes syntax highlighting in emacs
    {Char =< 0'9},
    {N1 is N0 * 10 + (Char - 0'0)},
    ( number(N1, N)
    ; {N = N1}
    ).

conditionalEat(Pred, [H|Tail], NTail) :-
    applyFun(Pred, [H]), !, conditionalEat(Pred, Tail, NTail).
conditionalEat(_, X, X) :- !.

whitespace --> eatWhitespace,
    ( (lineComment ; blockComment), !, whitespace
    ; []
    ).

eatWhitespace(In, Out) :- conditionalEat(isWhitespace, In, Out).

lineComment(In, Out) :- dash(In, InOut1), dash(InOut1, InOut2),
    dropWhile(isNotNl, InOut2, Out).

blockComment(In, Out) :- ocomment(In, InOut1),
    dropWhile(isNotDash, InOut1, InOut2), ccomment(InOut2, Out).

ws(Fun) --> whitespace, callFun(Fun).

isWhitespace(X) :- char_type(X, space).

isAlnum(X)             :- char_type(X, alnum).
isAlpha(X)             :- char_type(X, alpha).
isFSym(X)              :- char_type(X, csym).
isFSym(X)              :- char_code('\'', X).
isFSymf(X, [X|XS], XS) :- char_type(X, csymf).
isDQuote(X)            :- char_code('\"', X).
isBackslash(X)         :- char_code('\\', X).
isNotNl(X)             :- not(char_code('\n', X)).
isNotDash(X)           :- not(char_code('-', X)).
isNotEoc(X)            :- not(char_code('>', X)).

comma      --> ",".
bar        --> "|".
ampersand  --> "&".
lparen     --> "(".
rparen     --> ")".
lsqbracket --> "[".
rsqbracket --> "]".
lcubracket --> "{".
rcubracket --> "}".
lanbracket --> "<".
ranbracket --> ">".
dot        --> ".".
dots       --> dot, dot.
dash       --> "-".
colon      --> ":".
semicolon  --> ";".
bang       --> "!".
at         --> "@".
asterisk   --> "*".
plus       --> "+".
minus      --> dash.
dollar     --> "$".
question   --> "?".
equals     --> "=".
caret      --> "^".
slash      --> "/".
backslash  --> "\\".
hash       --> "#".
quote      --> "\'".
dquote     --> "\"".
oquote     --> "`".
tilde      --> "~".
usefeature --> oquote.
ocomment   --> lcubracket, dash.
ccomment   --> dash, rcubracket.
module     --> "module".
import     --> "import".
hiding     --> "hiding".
qualified  --> "qualified".
extends    --> "extends".
as         --> "as".
where      --> "where".
all        --> ("all"  ; uForAll).
some       --> ("some" ; uThereExists).
no         --> ("no"   ; uThereDoesNotExist).
lone       --> "lone".
one        --> "one".
lTrue      --> "True".
lFalse     --> "False".
lNot       --> ("not"     ; tilde ; uNotSign ; uTildeOperator).
lAnd       --> ("and"     ; ampersand, ampersand ; uLogicalAnd).
lOr        --> ("or"      ; bar, bar ; uLogicalOr).
lXor       --> ("xor"     ; caret ; uXor ; uCircledPlus).
lImplies   --> ("implies" ; equals, ranbracket ; uRightwardsDoubleArrow
                          ; uRightTack).
lThen      --> ("then"    ; lImplies).
lElse      --> ("else"    ; semicolon).
lIff       --> ("iff"     ; lanbracket, equals, ranbracket
                          ; uLeftRightDoubleArrow).
in         --> (colon; uElementOf).
eq         --> equals, equals.
neq        --> (slash, equals ; uNotEqualTo).
less       --> lanbracket.
lesseq     --> (less, equals ; uLessThanOrEqualTo).
greater    --> ranbracket.
greatereq  --> (greater, equals ; uGreaterThanOrEqualTo).
subset     --> ("in" ; uSubsetOf).
restricted(X) :- member(X,
    ["", "extends", "where", "all", "some", "no", "lone", "one", "True", "False", "and", "or", "xor", "implies", "else", "iff", "import", "as"]).

% unicode characters
uForAll                --> [8704].
uThereExists           --> [8707].
uThereDoesNotExist     --> [8708].
uNotSign               --> [172].
uTildeOperator         --> [8764].
uLogicalAnd            --> [8743].
uLogicalOr             --> [8744].
uXor                   --> [8891].
uCircledPlus           --> [8853].
uRightwardsDoubleArrow --> [8658].
uRightTack             --> [8866].
uLeftRightDoubleArrow  --> [8660].
uElementOf             --> [8712].
uNotEqualTo            --> [8800].
uLessThanOrEqualTo     --> [8804].
uGreaterThanOrEqualTo  --> [8805].
uMultiplicationSign    --> [215].
uDivisionSign          --> [247].
uDivisionSlash         --> [8725].
uUnion                 --> [8746].
uSetMinus              --> [8726].
uIntersection          --> [8745].
uSubsetOf              --> [8834].