show(X,D)      :- spaces(D),writef(X),nl.
show(X,Args,D) :- spaces(D),writef(X,Args),nl.

sprint(S, nameExType(Name,Type)) :-
	swritef(S, "%w %w", [Type,Name]).

sprint(S, exType(I,Type)) :-
	swritef(S, "<%w %w>", [I,Type]).

sprint(S, card(_,Low,High)) :-
	swritef(S,"[%w..%w]",[Low, High]).

sprintFact(S,Node) :-
	fact(_,_,_,Node,Value),
	swritef(S,"%w: %w",[Node, Value]).

spaces(D) :- foreach(between(1,D,_),write('  ')).

%% Debugging output for parsed tree

pprint(featureDef(NameExType,Rest),D) :-
	sprint(NameS, NameExType),
	show("F: %w", [NameS], D),
	DS is D + 1,
	pprint(Rest,DS),
	fail. %Fail - don't return anything

pprint(featureBody(Constraints,Rest),D) :-
	show("Feature Body",D),
	DS is D + 1,
	pprint(Constraints,DS),
	pprint(Rest,DS).

pprint(inlineFeatureDef(NameExType,Rest),D) :-
	sprint(NameS, NameExType),
	show("Inline feature def: %w",[NameS],D),
	DS is D + 1,
	pprint(Rest,DS).

pprint(featureBodyPattern(Rest),D) :-
	show("Feature Body Pattern",D),
	pprint(Rest,D).

pprint(mapping(Name,Node),D) :-
	sprintFact(NodeS,Node),
	show("Mapping: %w [%w]",[Name,NodeS],D).

pprint(mapping(Name,Node,Rest),D) :-
	sprintFact(NodeS,Node),
	show("Mapping /w rest: %w [%w]",[Name,NodeS],D),
	DS is D + 1,
	pprint(Rest,DS).

%Ignoring Constraints
pprint(subfeature(Card,_,Rest),D) :-
	sprint(CardS,Card),
	show("Subfeature %w",[CardS],D),
	DS is D + 1,
	%No constraints in Applet model
	%pprint(Constraints,DS),
	pprint(Rest,DS).

pprint(subfeatures(Rest),D) :-
	show("subfeatures",D),
	DS is D + 1,
	pprint(Rest,DS).

%% AST rewriting and pretty printing

pprint([],_).
pprint(List,D) :-
	is_list(List),
	DS is D + 1,
	foreach(member(C,List),pprint(C,DS)).
pprint(sleaf,_).
pprint(sfeature(Name,Value,Children),D) :-
	spaces(D),
	writef("%w = %w", [Name, Value]),
	nl,
	pprint(Children,D).
pprint(sinstances(_,Children),D) :-
	pprint(Children,D).

pprint(sfm(Feature)) :- pprint(Feature,0).

ast(featureDef(NameExType,Rest),Tree) :-
	attributes(NameExType,Name,Type),
	children(Rest,Children),
	ast(Children,Subtrees),
	Root =.. [sfeature, Name, Type, Subtrees],
	Tree =.. [sfm, Root],!.

ast(subfeature(_,_,Rest),Tree) :-
	attributes(Rest,Name,Type),
	mappings(Rest,Mappings),
	findall(Subtree,(member(M,Mappings),ast(M,Name,Subtree)),Subtrees),
	Subtrees \== [],
	Tree =.. [sinstances, Type, Subtrees],!.

%subfeature with no mapping, but its children has instances
ast(subfeature(_,_,Rest),Tree) :-
	mappings(Rest,[]),
	contains_submapping(Rest),
	children(Rest,Children),
	attributes(Rest,Name,Type),
	findall(Subtree, (member(C,Children),ast(C,Subtree)),Subtrees),
	Synthesized	=.. [sfeature, Name, Type, Subtrees],	
	Tree		=.. [sinstances, Type, [Synthesized]],!.

%otherwise, there is no instance of this feature
ast(subfeature(_,_,_),sleaf).

ast([],[sleaf]).
ast([H|T],Subtrees) :-
	ast(H,THead),
	ast(T,TTail),
	Subtrees = [THead|TTail],!. 

ast(mapping(_,Node,Rest),Name,Tree) :-
	mapped_value(Value,Node),
	ast(Rest,Subtrees),
	Tree =.. [sfeature, Name, Value, Subtrees].

contains_submapping(subfeature(_,_,Rest)) :-
	contains_submapping(Rest).
contains_submapping(X) :-
	mappings(X,Mappings),
	Mappings \== [].
contains_submapping(X) :-
	children(X,Cs),
	member(C,Cs),
	C \== X,
	contains_submapping(C).

children([],[]).
children([H|T],Result) :-
	List = [H|T],
	findall(L,(member(C,List),children(C,L)),Lists),
	append(Lists,Result),!.
children(featureBody(_,Rest),List) :-
	children(Rest,List),!.
children(featureBodyPattern(Rest),List) :-
	children(Rest,List),!.
children(mapping(_,_,Rest),List) :-
	children(Rest,List),!.
children(inlineFeatureDef(_,Rest),List) :-
	children(Rest,List),!.
children(X,List) :-
	X = subfeature(_,_,_), List = [X],!;
	X = subfeatures(Rest), List = Rest,!.
children(_,[]).
%	writef("Unrecognized children: %w", [X]),nl.

%Currently a copy of children, would be good to generalize
mappings([],[]).
mappings(List,Mappings) :-
	is_list(List),
	findall(L,(member(C,List),mappings(C,L)),Lists),
	append(Lists,Mappings),!.
mappings(inlineFeatureDef(_,Rest),Mappings) :-
	mappings(Rest,Mappings),!.
mappings(featureBody(_,Rest),Mappings) :-
	mappings(Rest,Mappings),!.
mappings(featureBodyPattern(Rest),Mappings) :-
	mappings(Rest,Mappings),!.
mappings(X,Mappings) :-
	X = mapping(_,_,_), Mappings = [X],!.
mappings(mapping(Name,Node),Mappings) :-
	Mapping =.. [mapping, Name, Node, []],
	Mappings = [Mapping],!.
mappings(_,[]).
%	writef("Unrecognized mappings: %w", [X]),nl.
	
constraints(featureBody(Constraints,_),Constraints).

attributes(inlineFeatureDef(NameExType,_),Name,Type) :-
	attributes(NameExType,Name,Type).
attributes(nameExType(Name,ExType),Name,Type) :-
	attributes(ExType,Type).

attributes(exType([],[]),true).
attributes(exType([],[type(fullyQualifiedName([],S))]),S).
attributes(X,X).

mapped_value(S,Node) :-
	fact(_,_,_,Node,(_,Value)),
	string_to_atom(S,Value),!.
mapped_value(true,Node) :-
	fact(_,_,_,Node,empty),!.
mapped_value(S,Node) :-
	fact(_,_,_,Node,Value),
	string_to_atom(S,Value).


